
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.awt.geom.*;
import java.math.*;


/**This class defines the spherical measures which go into
   the energy bound.  This bounds the energy of all configurations
   in a block based on the energy of the vertex configs and the
   error term. These are described in the part of the
   monograph dealing with the energy estimate (Lemma E).*/

public class SphericalMeasures {


    /**Gets the vertices of a box and then does inverse 
       stereographic projection.
       The boxes of type 2 corresponds to 
       the infinite point in the plane, which
       in turn corresponds to the point 
       (0,0,1) on the unit sphere.  **/

    public static Vector[] sphereVectors(Box B) {
	if(B.type==2) {
	    Vector[] V={new Vector(0,0,1)};
	    return(V);
	}
	Complex[] Z=B.toVertices();
	Vector[] V=new Vector[Z.length];
	for(int i=0;i<Z.length;++i) V[i]=Vector.inverseStereo(Z[i]);
	return(V);
    }

    /**This is the square of the diameter of the convex hull of the
       vertices of the box. It is used in our energy estimate*/

    public static double hullDiameterSquared(Box B) {
	Vector[] V=sphereVectors(B);
	if(V.length==1) return(0);
	double max=0;
	for(int i=0;i<V.length;++i) {
	    for(int j=i+1;j<V.length;++j) {
		double test=Vector.distSquared(V[i],V[j]);
		if(max<test) max=test;
	    }
	}
	return max;
    }

    /**This is the hull separation constant*/

    public static double hullSepConstant(Box B) {
	if(B.type==0) return hullSepConstant0(B);
	if(B.type==1) return hullSepConstant1(B);
	return 0;	 //point at infinity
    }

    /**this is the routine for squares*/

    public static double hullSepConstant0(Box B) {
	Vector[] V=sphereVectors(B);
	double[] s=new double[4];
	for(int i=0;i<4;++i) {
	    int j=(i+1)%4;
	    double d2=Vector.distSquared(V[i],V[j]);
	    s[i]=d2/4 + d2*d2/4;
	}
	double s02=Math.max(s[0],s[2]);
	double s13=Math.max(s[1],s[3]);
	return s02+s13;
    }
    
    /**this is the routine for segments*/
	
    public static double hullSepConstant1(Box B) {
	Vector[] V=sphereVectors(B);
	double d2=Vector.distSquared(V[0],V[1]);
	double s=d2/8 + d2*d2/32;
	return s;
    }





    
    public static double dotMaxUpper(Box B1,Box B2) {
	Vector[] V1=sphereVectors(B1);
	Vector[] V2=sphereVectors(B2);
	double max=-1;
	for(int i=0;i<V1.length;++i) {
	    for(int j=0;j<V2.length;++j) {
		double test=Vector.dot(V1[i],V2[j]);
		if(max<test) max=test;
	    }
	}

	if(B1.type==2) return max;
	if(B2.type==2) return max;
	double d1=hullSepConstant(B1);
	double d2=hullSepConstant(B2);
	max=max+d1+d2+d1*d2;
	return max;
    }

}
